#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <lsapi/lsapi.h>
#include "lsxcommand.h"
#include "hash_table.h"

#include "ModuleLog.h"

/*

  Search and Alias Data

*/

struct hash_table *tSearchEngines = NULL;
char *sSearchDefault = NULL;
HASH hSearchDefault;
HMENU hSearchMenu = NULL;

struct hash_table *tAliases = NULL;
HMENU hAliasMenu = NULL;


/*

*/

/***********************************************************
* void Search_SetDefault()                                 *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HASH hash;                                          *
*      The hash value for the default search engine token  *
*    - const char *key;                                    *
*      The KeyCase string token for the default engine.    *
*                         *  *  *  *                       *
* Allocates space for and stores the default search engine *
* token to look up in the hash table when needed.  This    *
* will also free the previous default if any was set.      *
***********************************************************/
static void Search_SetDefault(HASH hash, const char *key)
{
  char *NewDef = malloc((strlen(key) + 1) * sizeof(*NewDef));
  if (NewDef)
  {
    strcpy(NewDef, key);
    free(sSearchDefault);
    hSearchDefault = hash;
    sSearchDefault = NewDef;
  }
}

/***********************************************************
* BOOL LS_fgets()                                          *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - FILE *File;                                         *
*      A pointer to the FILE structure for the file.       *
*    - LPCSTR Key;                                         *
*      Corresponds to the key for LCReadNextConfig, this   *
*      argument is ignored.                                *
*    - LPSTR Buffer;                                       *
*      The buffer to store the line in.                    *
*    - size_t Size;                                        *
*      The size of Buffer in characters.                   *
*                         *  *  *  *                       *
* Provides an adapter between LCReadNextConfig's signature *
* and behavior and that of fgets.  This reads a line from  *
* the file and strips the trailing newline, when present.  *
***********************************************************/
static BOOL LS_fgets(FILE *File, LPCSTR Key, LPSTR Buffer, size_t Size)
{
  if (fgets(Buffer, Size, File) != NULL)
  {
    Size = strlen(Buffer);
    /* remove trailing space (unless we sliced a line) */ 
    if (Buffer[Size-1] == '\n')
      Buffer[Size-1] = 0;
    return TRUE;
  }
  return FALSE;
}

/***********************************************************
* void Config_Load()                                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *Source;                                 *
*      The source for this configuration type.  If the     *
*      string starts with a *, it is interpreted as a      *
*      step.rc setting, otherwise it's interpreted as a    *
*      file name.                                          *
*    - struct hash_table **pTable;                         *
*      Pointer to the pointer that holds the hash table    *
*      for this configuration type.  pTable may not be     *
*      NULL, but *pTable may. (it will create a new table, *
*      or build on an existing one)                        *
*    - HMENU *pMenu;                                       *
*      Pointer to the HMENU to store the generated menu in *
*      pMenu should not be NULL, *pMenu can be.  (any      *
*      existing menu will be destroyed and replaced.)      *
*    - UINT idWhat;                                        *
*      Identifies what config it is reading and serves as  *
*      a flag for menu ids.  This should be HMI_USER_ALIAS *
*      or HMI_USER_SEARCH.                                 *
*                         *  *  *  *                       *
* Parses a configuration, building a hash table and a      *
* heirarchical menu.                                       *
***********************************************************/
static void Config_Load(const char *Source, struct hash_table **pTable, HMENU *pMenu, UINT idWhat)
{
#define OPEN_BRACE '{'
#define CLOSE_BRACE '}'
#define OPEN_BRACE_STRING "{"
#define CLOSE_BRACE_STRING "}"
  char Line[MAX_LINE_LENGTH];
  const char *p;
  char OrigCase[MAX_LINE_LENGTH];
  char KeyCase[MAX_LINE_LENGTH];
  unsigned nOrigLength, nKeyLength;
  HASH hash;
  FILE *f;
#define MENU_STACK_SIZE 8
  /* 8 levels deep should be enough for anyone... */ 
  HMENU hMenuStack[MENU_STACK_SIZE];
  unsigned iMenuStackPtr = 0;
  UINT id = idWhat;
  unsigned nLine = 0, nEntries = 0;
  unsigned nSkip;
  const char *sWhat;

  BOOL (*pGetLine)(FILE *, LPCSTR, LPSTR, size_t);

  if (!Source)
    return;

  switch (idWhat)
  {
  case HMI_USER_SEARCH: sWhat = "search engines"; break;
  case HMI_USER_ALIAS: sWhat = "aliases"; break;
  }

  if (*Source == '*')
  {
    pGetLine = LCReadNextConfig;
    nSkip = strlen(Source); /* because LSAPI gives us junk at the beginning */ 
    f = LCOpen(NULL);
  }
  else
  {
    pGetLine = LS_fgets;
    nSkip = 0;
    f = fopen(Source, "rt");
    if (!f)
    {
      Complain("Unable to open file: %1%nwhile reading %2", Source, sWhat);
      return;
    }
  }

  hMenuStack[iMenuStackPtr] = CreatePopupMenu();

  /* file must end with a newline */ 
  while (pGetLine(f, Source, Line, MAX_LINE_LENGTH))
  {
    ++nLine;
    p = Parse_SkipSpace(Line + nSkip);

    /* ignore blank and comment lines */ 
    if (*p == 0 || *p == ';') continue;

    nOrigLength = nKeyLength = MAX_LINE_LENGTH;
    if (!Parse_GetHashedToken(p, &p, OrigCase, &nOrigLength, &hash, KeyCase, &nKeyLength))
      continue;

    if (p) p = Parse_SkipSpace(p);

    if (*OrigCase == CLOSE_BRACE)
    {
      /* end group */ 
      if (iMenuStackPtr > 0)
      {
        --iMenuStackPtr;
      }
      else
      {
        Complain("Too many '" CLOSE_BRACE_STRING "'s found.  (#%1!d!)%nwhile reading %2", nLine, sWhat);
      }
    }
    else if (p && *p == OPEN_BRACE)
    {
      /* begin group */ 
      if (iMenuStackPtr < MENU_STACK_SIZE)
      {
        ++iMenuStackPtr;
        hMenuStack[iMenuStackPtr] = CreatePopupMenu();
        AppendMenu(hMenuStack[iMenuStackPtr-1], MF_STRING|MF_POPUP, (UINT)hMenuStack[iMenuStackPtr], OrigCase);
      }
      else
      {
        Complain("Categories are nested too deeply (#%1!d!):%n%2%nwhile reading %3", nLine, Line, sWhat);
      }
    }
#define HASH_OF_DEFAULT 0xE4DF46D5
    else if (idWhat==HMI_USER_SEARCH && hash==HASH_OF_DEFAULT && strcmp(KeyCase, "default")==0)
    {
      if (Parse_GetHashedToken(p, &p, OrigCase, NULL, &hash, KeyCase, NULL))
        Search_SetDefault(hash, KeyCase);
    }
#define HASH_OF_STOP 0x930CA7F2
    else if (hash==HASH_OF_STOP && strcmp(KeyCase, "stop")==0 && (!p || !*p))
    {
      /* end processing */ 
      break;
    }
    else
    {
      *pTable = HashTable_Set(*pTable, hash, KeyCase, p);
      /* put a mark next to the default, but only if the default is specified earlier in the file */ 
      if (hash == hSearchDefault && sSearchDefault && strcmp(KeyCase, sSearchDefault)==0)
        AppendMenu(hMenuStack[iMenuStackPtr], MF_STRING|MF_CHECKED, ++id, OrigCase);
      else
        AppendMenu(hMenuStack[iMenuStackPtr], MF_STRING, ++id, OrigCase);
      ++nEntries;
    }
  }

  if (*Source == '*')
    LCClose(f);
  else
    fclose(f);

  if (nEntries == 0)
  {
    DestroyMenu(hMenuStack[0]);
    hMenuStack[0] = NULL;
  }

  *pMenu = hMenuStack[0];

#ifndef NLOG
  HashTable_Dump(*pTable);
#endif
}

/***********************************************************
* void Search_Load()                                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Reads the search engine list and builds the hash table   *
* and submenu for it.                                      *
***********************************************************/
void Search_Load(struct CommandSettings *cs)
{
  TRACE1("Search_Load");
  ASSERT(READABLE(cs));
  Config_Load(cs->SearchEngineList, &tSearchEngines, &hSearchMenu, HMI_USER_SEARCH);
}

/***********************************************************
* void Alias_Load()                                        *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Reads the alias list and builds the hash table and       *
* submenu for it.                                          *
***********************************************************/
void Alias_Load(struct CommandSettings *cs)
{
  TRACE1("Alias_Load");
  Config_Load("*CommandAlias", &tAliases, &hAliasMenu, HMI_USER_ALIAS);
}

/***********************************************************
* void Search_Unload()                                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Releases the resources allocated in Search_Load()        *
***********************************************************/
void Search_Unload(struct CommandSettings *cs)
{
  TRACE1("Search_Unload");
  tSearchEngines = HashTable_Destroy(tSearchEngines);
  free(sSearchDefault);
  sSearchDefault = NULL;
  DestroyMenu(hSearchMenu);
  hSearchMenu = NULL;
}

/***********************************************************
* void Alias_Unload()                                      *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Releases the resources allocated in Alias_Load()         *
***********************************************************/
void Alias_Unload(struct CommandSettings *cs)
{
  TRACE1("Alias_Unload");
  tAliases = HashTable_Destroy(tAliases);
  DestroyMenu(hAliasMenu);
  hAliasMenu = NULL;
}

/***********************************************************
* void Search_Execute()                                    *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND hWnd;                                          *
*      The window handle to be set to LSExecuteEx.         *
*    - HASH hash;                                          *
*      The hash for the search engine token.  This is      *
*      ignored if no key is given.                         *
*    - const char *key;                                    *
*      The KeyCase token representing the search engine to *
*      use, or NULL or an empty string to use the default. *
*      This must not include the leading ? unless it's     *
*      really a part of the search engine token.           *
*    - const char *args;                                   *
*      The arguments to pass to the search engine.         *
*                         *  *  *  *                       *
* Looks up the search engine, assembles the final URL, and *
* executes it.                                             *
***********************************************************/
extern struct CommandSettings *cs;
void Search_Execute(HWND hWnd, HASH hash, const char *key, const char *args)
{
  /* key should NOT include the '?'     */ 
  /* the hash would be wrong if it did. */ 

  unsigned nLength;
  char *URL;
  const char *Template;
  int hError;

  /* check for validity */ 

  if (!tSearchEngines)
  {
    Complain("No search engines defined.");
    return;
  }

  if (!args || !*args)
  {
    Complain("Nothing to search for.");
    return;
  }

  if (!key || !*key)
  {
    /* no specific engine requested: use default */ 
    if (sSearchDefault)
    {
      hash = hSearchDefault;
      key = sSearchDefault;
    }
    else
    {
      Complain("No default search engine specified.");
      return;
    }
  }

  /* lookup search engine */ 
  Template = HashTable_Get(tSearchEngines, hash, key);
  if (!Template)
  {
    Complain("Unable to find search engine: %1", key);
    return;
  }

  /* get space and expand URL */ 
  nLength = 3*strlen(args) + strlen(Template) + 1;
  URL = malloc(nLength*sizeof(*URL));
  if (URL)
  {
    if (Parse_ReplaceSearchArgs(Template, args, URL, nLength))
    {
      /* go - URLs must use "open" verb if not using a specific program */ 
      if (*cs->BrowserPath)
        hError = (int)LSExecuteEx(hWnd, "open", cs->BrowserPath, URL,  NULL, SW_NORMAL);
      else
        hError = (int)LSExecuteEx(hWnd, "open", URL,             NULL, NULL, SW_NORMAL);
      ComplainExecError(hError, NULL, URL);
    }
    free(URL);
  }
}

/***********************************************************
* BOOL Alias_Execute()                                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HASH hash;                                          *
*      The hash of the alias token.                        *
*    - const char *key;                                    *
*      The KeyCase version of the alias token.             *
*    - const char *args;                                   *
*      Any arguments to be passed to the alias, or NULL or *
*      an empty string if there are none.                  *
*                         *  *  *  *                       *
* Looks up an alias and executes it.  Returns TRUE if a    *
* matching alias was found.  Beware of recursion.          *
***********************************************************/
BOOL Alias_Execute(HASH hash, const char *key, const char *args)
{
  const char *sValue;
  unsigned nValueLength, nLength;
  char *sCommand;
  BOOL bOk = TRUE;

  if (!tAliases)
    return FALSE;

  sValue = HashTable_Get(tAliases, hash, key);
  if (!sValue)
    return FALSE;

  nValueLength = strlen(sValue);
  nLength = nValueLength + (args ? strlen(args) + 2 : 1);
  sCommand = malloc(nLength * sizeof(*sCommand));
  if (sCommand)
  {
    if (strstr(sValue, "<args>"))
    {
      if (args && *args)
        Parse_ReplaceArgs(sValue, args, sCommand, nLength);
      else
      {
        Complain("Alias '%1' requires arguments.", key);
        bOk = FALSE;
      }
    }
    else
    {
      strcpy(sCommand, sValue);
      if (args && *args)
      {
        *(sCommand + nValueLength) = ' ';
        strcpy(sCommand + nValueLength + 1, args);
      }
    }
    if (bOk)
      ExecCommand(sCommand, TRUE);
    free(sCommand);
  }

  return TRUE;
}
